MODULE trcadv_muscl2
   !!==============================================================================
   !!                       ***  MODULE  trcadv_muscl2  ***
   !! Ocean passive tracers:  horizontal & vertical advective trend
   !!==============================================================================
#if defined key_passivetrc
   !!----------------------------------------------------------------------
   !!   tra_adv_muscl2 : update the tracer trend with the horizontal
   !!                    and vertical advection trends using MUSCL2 scheme
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce_trc         ! ocean dynamics and active tracers variables
   USE trc             ! ocean passive tracers variables
   USE lbclnk          ! ocean lateral boundary conditions (or mpp link)

   IMPLICIT NONE
   PRIVATE

   !! * Accessibility
   PUBLIC trc_adv_muscl2        ! routine called by trcstp.F90

   !! * Module variable
   REAL(wp), DIMENSION(jpk) ::   &
      rdttrc                     ! vertical profile of tracer time-step

   !! * Substitutions
#  include "passivetrc_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LODYC-IPSL (2003)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE trc_adv_muscl2( kt )
      !!----------------------------------------------------------------------
      !!                   ***  ROUTINE trc_adv_muscl2  ***
      !!
      !! ** Purpose :   Compute the now trend due to total advection of passi-
      !!      ve tracer using a MUSCL scheme (Monotone Upstream-
      !!      Centered Scheme for Conservation Laws) and add it to the general
      !!      tracer trend.
      !!
      !! ** Method : MUSCL scheme plus centered scheme at ocean boundaries
      !!
      !! ** Action : - update tra with the now advective tracer trends
      !!             - save trends in trtrd ('key_trc_diatrd')
      !!
      !! References :                
      !!      Estubier, A., and M. Levy, Notes Techn. Pole de Modelisation
      !!	IPSL, Sept. 2000 (http://www.lodyc.jussieu.fr/opa)
      !!
      !! History :
      !!        !  06-00  (A.Estublier)  for passive tracers
      !!   9.0  !  03-04  (C. Ethe, G. Madec)  F90: Free form and module
      !!----------------------------------------------------------------------
      !! * modules used
#if defined key_trcbbl_adv
      USE oce_trc            , zun => ua,  &  ! use ua as workspace
         &                     zvn => va      ! use va as workspace
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: zwn
#else
      USE oce_trc            , zun => un,  &  ! When no bbl, zun == un
                               zvn => vn,  &  !              zvn == vn
                               zwn => wn      !              zwn == wn
#endif
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt         ! ocean time-step

      !! * Local declarations
      INTEGER ::   ji, jj, jk,jn               ! dummy loop indices
      REAL(wp), DIMENSION (jpi,jpj,jpk) ::   &
         zt1, zt2, ztp1, ztp2
      REAL(wp) ::   zu, zv, zw, zeu, zev, zew, zbtr, ztra
      REAL(wp) ::   z0u, z0v, z0w
      REAL(wp) ::   zzt1, zzt2, zalpha

#if defined key_trc_diatrd
      REAL(wp) ::   ztai, ztaj
      REAL(wp) ::   zfui, zfvj
#endif
      !!----------------------------------------------------------------------

      IF( kt == nittrc000 .AND. lwp ) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'trc_adv_muscl2 : MUSCL2 advection scheme'
         WRITE(numout,*) '~~~~~~~~~~~~~~~'
         rdttrc(:) =  rdttra(:) * FLOAT(ndttrc)
      ENDIF
      
#if defined key_trcbbl_adv        
      ! Advective bottom boundary layer
      ! -------------------------------
      zun(:,:,:) = un (:,:,:) - u_trc_bbl(:,:,:)
      zvn(:,:,:) = vn (:,:,:) - v_trc_bbl(:,:,:)
      zwn(:,:,:) = wn (:,:,:) + w_trc_bbl( :,:,:)
#endif


      DO jn = 1, jptra

         ! I. Horizontal advective fluxes
         ! ------------------------------

         ! first guess of the slopes
         ! interior values
         DO jk = 1, jpkm1
            DO jj = 1, jpjm1      
               DO ji = 1, fs_jpim1   ! vector opt.
                  zt1(ji,jj,jk) = umask(ji,jj,jk) * ( trb(ji+1,jj,jk,jn) - trb(ji,jj,jk,jn) )
                  zt2(ji,jj,jk) = vmask(ji,jj,jk) * ( trb(ji,jj+1,jk,jn) - trb(ji,jj,jk,jn) )
               END DO
            END DO
         END DO
         ! bottom values
         zt1(:,:,jpk) = 0.e0
         zt2(:,:,jpk) = 0.e0

         ! lateral boundary conditions on zt1, zt2  (changed sign)
         CALL lbc_lnk( zt1, 'U', -1. )
         CALL lbc_lnk( zt2, 'V', -1. )

         ! Slopes
         ! interior values
         DO jk = 1, jpkm1
            DO jj = 2, jpj
               DO ji = fs_2, jpi   ! vector opt.
                  ztp1(ji,jj,jk) =                    ( zt1(ji,jj,jk) + zt1(ji-1,jj  ,jk) )   &
                     &           * ( 0.25 + SIGN( 0.25, zt1(ji,jj,jk) * zt1(ji-1,jj  ,jk) ) )
                  ztp2(ji,jj,jk) =                    ( zt2(ji,jj,jk) + zt2(ji  ,jj-1,jk) )   &
                     &           * ( 0.25 + SIGN( 0.25, zt2(ji,jj,jk) * zt2(ji  ,jj-1,jk) ) )
               END DO
            END DO
         END DO
         ! bottom values
         ztp1(:,:,jpk) = 0.e0 
         ztp2(:,:,jpk) = 0.e0

         ! Slopes limitation
         DO jk = 1, jpkm1
            DO jj = 2, jpj
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  ztp1(ji,jj,jk) = SIGN( 1., ztp1(ji,jj,jk) )   &
                     &           * MIN(    ABS( ztp1(ji  ,jj,jk) ),   &
                     &                  2.*ABS( zt1 (ji-1,jj,jk) ),   &
                     &                  2.*ABS( zt1 (ji  ,jj,jk) ) )

                  ztp2(ji,jj,jk) = SIGN( 1., ztp2(ji,jj,jk) )   &
                     &           * MIN(    ABS( ztp2(ji,jj  ,jk) ),   &
                     &                  2.*ABS( zt2 (ji,jj-1,jk) ),   &
                     &                  2.*ABS( zt2 (ji,jj  ,jk) ) )
               END DO
            END DO
         END DO

         ! Advection terms
         ! interior values
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1      
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  ! volume fluxes
#if defined key_s_coord || defined key_partial_steps
                  zeu = e2u(ji,jj) * fse3u(ji,jj,jk) * zun(ji,jj,jk)
                  zev = e1v(ji,jj) * fse3v(ji,jj,jk) * zvn(ji,jj,jk)
#else
                  zeu = e2u(ji,jj) * zun(ji,jj,jk)
                  zev = e1v(ji,jj) * zvn(ji,jj,jk)
#endif
                  ! MUSCL fluxes
                  z0u = SIGN( 0.5, zun(ji,jj,jk) )            
                  zalpha = 0.5 - z0u
                  zu  = z0u - 0.5 * zun(ji,jj,jk) * rdttrc(jk) / e1u(ji,jj)
                  zzt1 = trb(ji+1,jj,jk,jn) + zu*ztp1(ji+1,jj,jk)
                  zzt2 = trb(ji  ,jj,jk,jn) + zu*ztp1(ji  ,jj,jk)
                  zt1(ji,jj,jk) = zeu * ( zalpha * zzt1 + (1.-zalpha) * zzt2 )

                  z0v = SIGN( 0.5, zvn(ji,jj,jk) )            
                  zalpha = 0.5 - z0v
                  zv  = z0v - 0.5 * zvn(ji,jj,jk) * rdttrc(jk) / e2v(ji,jj)
                  zzt1 = trb(ji,jj+1,jk,jn) + zv*ztp2(ji,jj+1,jk)
                  zzt2 = trb(ji,jj  ,jk,jn) + zv*ztp2(ji,jj  ,jk)
                  zt2(ji,jj,jk) = zev * ( zalpha * zzt1 + (1.-zalpha) * zzt2 )

               END DO
            END DO
         END DO


         DO jk = 1, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
#if defined key_s_coord || defined key_partial_steps
                  zev = e1v(ji,jj) * fse3v(ji,jj,jk)
                  IF( umask(ji,jj,jk) == 0. ) THEN
                     IF( zun(ji+1,jj,jk) > 0. .AND. ji /= jpi ) THEN
                        zt1(ji+1,jj,jk) = e2u(ji+1,jj)* fse3u(ji+1,jj,jk)   &
                           &            * zun(ji+1,jj,jk) * ( trb(ji+1,jj,jk,jn) + trb(ji+2,jj,jk,jn) ) * 0.5
                     ENDIF
                     IF( zun(ji-1,jj,jk) < 0. ) THEN
                        zt1(ji-1,jj,jk) = e2u(ji-1,jj)* fse3u(ji-1,jj,jk)   &
                           &            * zun(ji-1,jj,jk) * ( trb(ji-1,jj,jk,jn) + trb(ji  ,jj,jk,jn) ) * 0.5
                     ENDIF
                  ENDIF
                  IF( vmask(ji,jj,jk) == 0. ) THEN
                     IF( zvn(ji,jj+1,jk) > 0. .AND. jj /= jpj ) THEN
                        zt2(ji,jj+1,jk) = e1v(ji,jj+1) * fse3v(ji,jj+1,jk)   &
                           &            * zvn(ji,jj+1,jk) * ( trb(ji,jj+1,jk,jn) + trb(ji,jj+2,jk,jn) ) * 0.5
                     ENDIF
                     IF( zvn(ji,jj-1,jk) < 0. ) THEN
                        zt2(ji,jj-1,jk) = e1v(ji,jj-1)* fse3v(ji,jj-1,jk)   &
                           &            * zvn(ji,jj-1,jk) * ( trb(ji,jj-1,jk,jn) + trb(ji  ,jj,jk,jn) ) * 0.5
                     ENDIF
                  ENDIF

#else
                  IF( umask(ji,jj,jk) == 0. ) THEN
                     IF( zun(ji+1,jj,jk) > 0. .AND. ji /= jpi ) THEN
                        zt1(ji+1,jj,jk) = e2u(ji+1,jj) * zun(ji+1,jj,jk) * ( trb(ji+1,jj,jk,jn) + trb(ji+2,jj,jk,jn) ) * 0.5
                     ENDIF
                     IF( zun(ji-1,jj,jk) < 0. ) THEN
                        zt1(ji-1,jj,jk) = e2u(ji-1,jj) * zun(ji-1,jj,jk) * ( trb(ji-1,jj,jk,jn) + trb(ji  ,jj,jk,jn) ) * 0.5
                     ENDIF
                  ENDIF
                  IF( vmask(ji,jj,jk) == 0. ) THEN
                     IF( zvn(ji,jj+1,jk) > 0. .AND. jj /= jpj ) THEN
                        zt2(ji,jj+1,jk) = e1v(ji,jj+1) * zvn(ji,jj+1,jk) * ( trb(ji,jj+1,jk,jn) + trb(ji,jj+2,jk,jn) ) * 0.5
                     ENDIF
                     IF( zvn(ji,jj-1,jk) < 0. ) THEN
                        zt2(ji,jj-1,jk) = e1v(ji,jj-1) * zvn(ji,jj-1,jk) * ( trb(ji,jj-1,jk,jn) + trb(ji  ,jj,jk,jn) ) * 0.5
                     ENDIF
                  ENDIF
#endif
               END DO
            END DO
         END DO

         ! lateral boundary conditions on zt1, zt2   (changed sign)
         CALL lbc_lnk( zt1, 'U', -1. )
         CALL lbc_lnk( zt2, 'V', -1. )

         ! Compute and add the horizontal advective trend

         DO jk = 1, jpkm1
            DO jj = 2, jpjm1      
               DO ji = fs_2, fs_jpim1   ! vector opt.
#if defined key_s_coord || defined key_partial_steps
                  zbtr = 1. / ( e1t(ji,jj)*e2t(ji,jj)*fse3t(ji,jj,jk) )
#else
                  zbtr = 1. / ( e1t(ji,jj)*e2t(ji,jj) )
#endif
                  ! horizontal advective trends
                  ztra = - zbtr * ( zt1(ji,jj,jk) - zt1(ji-1,jj  ,jk  )   &
                     &            + zt2(ji,jj,jk) - zt2(ji  ,jj-1,jk  ) )
                  ! add it to the general tracer trends
                  tra(ji,jj,jk,jn) = tra(ji,jj,jk,jn) + ztra
#if defined key_trc_diatrd
                  ! recompute the trends in i- and j-direction as Uh gradh(T)
#   if defined key_s_coord || defined key_partial_steps
                  zfui =  e2u(ji  ,jj) * fse3u(ji,  jj,jk) * un(ji,  jj,jk)   &
                     & -  e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * un(ji-1,jj,jk)
                  zfvj =  e1v(ji,jj  ) * fse3v(ji,jj  ,jk) * vn(ji,jj  ,jk)   &
                     & -  e1v(ji,jj-1) * fse3v(ji,jj-1,jk) * vn(ji,jj-1,jk)
#   else
                  zfui = e2u(ji  ,jj) * un(ji,  jj,jk)   &
                     & - e2u(ji-1,jj) * un(ji-1,jj,jk)
                  zfvj = e1v(ji,jj  ) * vn(ji,jj  ,jk)   &
                     & - e1v(ji,jj-1) * vn(ji,jj-1,jk)
#   endif
                  ztai =-zbtr * (  zt1(ji,jj,jk) - zt1(ji-1,jj  ,jk) - trn(ji,jj,jk,jn) * zfui  )
                  ztaj =-zbtr * (  zt2(ji,jj,jk) - zt2(ji  ,jj-1,jk) - trn(ji,jj,jk,jn) * zfvj  )
                  ! save i- and j- advective trends computed as Uh gradh(T)
                  trtrd(ji,jj,jk,jn,1) = ztai
                  trtrd(ji,jj,jk,jn,2) = ztaj

#endif
               END DO
            END DO
         END DO

         IF(l_ctl) THEN         ! print mean trends (used for debugging)
            ztra = SUM( tra(2:nictl,2:njctl,1:jpkm1,jn) * tmask(2:nictl,2:njctl,1:jpkm1) )
            WRITE(numout,*) ' trc/had  - ',ctrcnm(jn),' : ', ztra-tra_ctl(jn), ' muscl' 
            tra_ctl(jn) = ztra 
         ENDIF

         ! II. Vertical advective fluxes
         ! -----------------------------

         ! First guess of the slope
         ! interior values
         DO jk = 2, jpkm1
            zt1(:,:,jk) = tmask(:,:,jk) * ( trb(:,:,jk-1,jn) - trb(:,:,jk,jn) )
         END DO
         ! surface and bottom boundary conditions
         zt1 (:,:, 1 ) = 0.e0 
         zt1 (:,:,jpk) = 0.e0

         ! Slopes
         DO jk = 2, jpkm1
            DO jj = 1, jpj
               DO ji = 1, jpi
                  ztp1(ji,jj,jk) =                    ( zt1(ji,jj,jk) + zt1(ji,jj,jk+1) )   &
                     &           * ( 0.25 + SIGN( 0.25, zt1(ji,jj,jk) * zt1(ji,jj,jk+1) ) )
               END DO
            END DO
         END DO

         ! Slopes limitation
         ! interior values
         DO jk = 2, jpkm1
            DO jj = 1, jpj
               DO ji = 1, jpi
                  ztp1(ji,jj,jk) = SIGN( 1., ztp1(ji,jj,jk) )   &
                     &           * MIN(    ABS( ztp1(ji,jj,jk  ) ),   &
                     &                  2.*ABS( zt1 (ji,jj,jk+1) ),   &
                     &                  2.*ABS( zt1 (ji,jj,jk  ) ) )
               END DO
            END DO
         END DO

         ! surface values
         ztp1(:,:,1) = 0. 

         ! vertical advective flux
         ! interior values
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1      
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zew = zwn(ji,jj,jk+1)
                  z0w = SIGN( 0.5, zwn(ji,jj,jk+1) )
                  zalpha = 0.5 + z0w
                  zw  = z0w - 0.5 * zwn(ji,jj,jk+1)* rdttrc(jk)/ fse3w(ji,jj,jk+1)
                  zzt1 = trb(ji,jj,jk+1,jn) + zw*ztp1(ji,jj,jk+1)
                  zzt2 = trb(ji,jj,jk  ,jn) + zw*ztp1(ji,jj,jk  )
                  zt1(ji,jj,jk+1) = zew * ( zalpha * zzt1 + (1.-zalpha)*zzt2 )
               END DO
            END DO
         END DO
         DO jk = 2, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  IF( tmask(ji,jj,jk+1) == 0. ) THEN
                     IF( zwn(ji,jj,jk) > 0. ) THEN
                        zt1(ji,jj,jk) = zwn(ji,jj,jk) * ( trb(ji,jj,jk-1,jn) + trb(ji,jj,jk,jn) ) * 0.5
                     ENDIF
                  ENDIF
               END DO
            END DO
         END DO

         ! surface values
         IF( lk_dynspg_fsc .OR. lk_dynspg_fsc_tsk ) THEN        ! free surface-constant volume
            zt1(:,:, 1 ) = zwn(:,:,1) * trb(:,:,1,jn)
         ELSE                                                   ! rigid lid : flux set to zero
            zt1(:,:, 1 ) = 0.e0
         ENDIF

         ! bottom values
         zt1(:,:,jpk) = 0.e0

         ! Compute & add the vertical advective trend

         DO jk = 1, jpkm1
            DO jj = 2, jpjm1      
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zbtr = 1. / fse3t(ji,jj,jk)
                  ! horizontal advective trends
                  ztra = - zbtr * ( zt1(ji,jj,jk) - zt1(ji,jj,jk+1) )
                  ! add it to the general tracer trends
                  tra(ji,jj,jk,jn) =  tra(ji,jj,jk,jn) + ztra
#if defined key_trc_diatrd
                  ! save the vertical advective trends computed as w gradz(T)
                  trtrd(ji,jj,jk,jn,3) = ztra - trn(ji,jj,jk,jn) * hdivn(ji,jj,jk)
#endif
               END DO
            END DO
         END DO

         IF(l_ctl) THEN         ! print mean trends (used for debugging)
            ztra = SUM( tra(2:nictl,2:njctl,1:jpkm1,jn) * tmask(2:nictl,2:njctl,1:jpkm1) )
            WRITE(numout,*) ' trc/zad  - ',ctrcnm(jn),' : ', ztra-tra_ctl(jn), ' muscl2' 
            tra_ctl(jn) = ztra 
         ENDIF

      END DO

   END SUBROUTINE trc_adv_muscl2

#else

   !!----------------------------------------------------------------------
   !!   Default option                                         Empty module
   !!----------------------------------------------------------------------
CONTAINS
   SUBROUTINE trc_adv_muscl2( kt ) 
      INTEGER, INTENT(in) :: kt
      WRITE(*,*) 'trc_adv_muscl2: You should not have seen this print! error?', kt
   END SUBROUTINE trc_adv_muscl2
#endif

   !!======================================================================
END MODULE trcadv_muscl2
